@@ -21,25 +21,27 @@ Can be used both for client and transport authentication
 See client_nb.py
 """
 
-from protocol import NS_SASL, NS_SESSION, NS_STREAMS, NS_BIND, NS_AUTH
-from protocol import NS_STREAM_MGMT
-from protocol import Node, NodeProcessed, isResultNode, Iq, Protocol, JID
-from plugin import PlugIn
-from smacks import Smacks
+from __future__ import unicode_literals
+
+from .protocol import NS_SASL, NS_SESSION, NS_STREAMS, NS_BIND, NS_AUTH
+from .protocol import NS_STREAM_MGMT
+from .protocol import Node, NodeProcessed, isResultNode, Iq, Protocol, JID
+from .plugin import PlugIn
+import sys
+import re
 import base64
-import random
-import itertools
-import dispatcher_nb
-import hashlib
+from . import dispatcher_nb
 import hmac
 import hashlib
 
 import logging
 log = logging.getLogger('nbxmpp.auth_nb')
 
+from . import rndg
+
 def HH(some): return hashlib.md5(some).hexdigest()
 def H(some): return hashlib.md5(some).digest()
-def C(some): return ':'.join(some)
+def C(some): return b':'.join(some)
 
 try:
     kerberos = __import__('kerberos')
@@ -55,9 +57,18 @@ SASL_SUCCESS = 'success'
 SASL_UNSUPPORTED = 'not-supported'
 SASL_IN_PROCESS = 'in-process'
 
-def challenge_splitter(data):
+# compile the search regex for _challenge_splitter
+_challenge_regex = re.compile("""
+    (\w+)      # keyword
+    =
+    ("[^"]+"|[^,]+)   # value
+    ,?         # optional comma separator
+""", re.VERBOSE)
+
+def _challenge_splitter(data):
     """
-    Helper function that creates a dict from challenge string
+    Helper function that creates a dict from challenge string. Used
+    for DIGEST-MD5 authentication.
 
     Sample challenge string:
       - username="example.org",realm="somerealm",
@@ -68,75 +79,50 @@ def challenge_splitter(data):
       - dict['qop'] = ('auth','auth-int','auth-conf')
       - dict['realm'] = 'somerealm'
     """
-    X_KEYWORD, X_VALUE, X_END = 0, 1, 2
-    quotes_open = False
-    keyword, value = '', ''
     dict_ = {}
-    arr = None
-
-    expecting = X_KEYWORD
-    for iter_ in range(len(data) + 1):
-        end = False
-        if iter_ == len(data):
-            expecting = X_END
-            end = True
-        else:
-            char = data[iter_]
-        if expecting == X_KEYWORD:
-            if char == '=':
-                expecting  = X_VALUE
-            elif char in (',', ' ', '\t'):
-                pass
-            else:
-                keyword = '%s%c' % (keyword, char)
-        elif expecting == X_VALUE:
-            if char == '"':
-                if quotes_open:
-                    end = True
-                else:
-                    quotes_open = True
-            elif char in (',', ' ', '\t'):
-                if quotes_open:
-                    if not arr:
-                        arr = [value]
-                    else:
-                        arr.append(value)
-                    value = ""
-                else:
-                    end = True
-            else:
-                value = '%s%c' % (value, char)
-        if end:
-            if arr:
-                arr.append(value)
-                dict_[keyword] = arr
-                arr = None
-            else:
-                dict_[keyword] = value
-            value, keyword = '', ''
-            expecting = X_KEYWORD
-            quotes_open = False
+    for match in _challenge_regex.finditer(data):
+        k = match.group(1)
+        v = match.group(2)
+        if v.startswith('"'):
+            v = v[1:-1] # Remove quote
+        if v.find(',') >= 0:
+            v = v.split(',') # Split using comma
+        dict_[k] = v
     return dict_
 
-def scram_parse(chatter):
+def _scram_parse(chatter):
+    """Helper function. Used for SCRAM-SHA-1, SCRAM-SHA-1-PLUS authentication"""
     return dict(s.split('=', 1) for s in chatter.split(','))
 
+SASL_AUTHENTICATION_MECHANISMS = \
+    set(['ANONYMOUS', 'EXTERNAL', 'GSSAPI', 'SCRAM-SHA-1-PLUS', 'SCRAM-SHA-1',
+         'DIGEST-MD5', 'PLAIN', 'X-MESSENGER-OAUTH2'])
+
 class SASL(PlugIn):
     """
     Implements SASL authentication. Can be plugged into NonBlockingClient
     to start authentication
     """
 
-    def __init__(self, username, password, on_sasl):
+    def __init__(self, username, password, on_sasl, channel_binding,
+                 auth_mechs):
         """
         :param username: XMPP username
         :param password: XMPP password
         :param on_sasl: Callback, will be called after each SASL auth-step.
+        :param channel_binding: TLS channel binding data, None if the 
+               binding data is not available
+        :param auth_mechs: Set of valid authentication mechanisms.
+               Possible entries are:
+               'ANONYMOUS', 'EXTERNAL', 'GSSAPI', 'SCRAM-SHA-1-PLUS',
+               'SCRAM-SHA-1', 'DIGEST-MD5', 'PLAIN', 'X-MESSENGER-OAUTH2'
         """
         PlugIn.__init__(self)
         self.username = username
         self.password = password
         self.on_sasl = on_sasl
+        self.channel_binding = channel_binding
+        self.enabled_auth_mechs = auth_mechs
         self.realm = None
 
     def plugin(self, owner):
@@ -173,7 +159,7 @@ class SASL(PlugIn):
         Start authentication. Result can be obtained via "SASL.startsasl"
         attribute and will be either SASL_SUCCESS or SASL_FAILURE
 
-        Note that successfull auth will take at least two Dispatcher.Process()
+        Note that successful auth will take at least two Dispatcher.Process()
         calls.
         """
         if self.startsasl:
@@ -196,10 +182,24 @@ class SASL(PlugIn):
             self.startsasl='not-supported'
             log.info('SASL not supported by server')
             return
-        self.mecs = []
-        for mec in feats.getTag('mechanisms', namespace=NS_SASL).getTags(
-        'mechanism'):
-            self.mecs.append(mec.getData())
+
+        self.mecs = set(
+            mec.getData() 
+            for mec 
+            in feats.getTag('mechanisms', namespace=NS_SASL).getTags('mechanism')
+        ) & self.enabled_auth_mechs
+       
+        # Password based authentication mechanism ordered by strength.
+        # If the server supports a mechanism disable all weaker mechanisms.
+        password_auth_mechs_strength = ['SCRAM-SHA-1-PLUS', 'SCRAM-SHA-1',
+            'DIGEST-MD5', 'PLAIN', 'X-MESSENGER-OAUTH2']
+        if self.channel_binding == None:
+            password_auth_mechs_strength.remove('SCRAM-SHA-1-PLUS')
+        for i in range(0, len(password_auth_mechs_strength)):
+            if password_auth_mechs_strength[i] in self.mecs:
+                for m in password_auth_mechs_strength[i + 1:]:
+                    self.mecs.discard(m)
+                break
 
         self._owner.RegisterHandler('challenge', self.SASLHandler,
             xmlns=NS_SASL)
@@ -218,9 +218,9 @@ class SASL(PlugIn):
             raise NodeProcessed
         if "EXTERNAL" in self.mecs:
             self.mecs.remove('EXTERNAL')
-            sasl_data = u'%s@%s' % (self.username, self._owner.Server)
-            sasl_data = sasl_data.encode('utf-8').encode('base64').replace(
-                '\n', '')
+            sasl_data = '%s@%s' % (self.username, self._owner.Server)
+            sasl_data = base64.b64encode(sasl_data.encode('utf-8')).decode(
+                'utf-8').replace('\n', '')
             node = Node('auth', attrs={'xmlns': NS_SASL,
                 'mechanism': 'EXTERNAL'}, payload=[sasl_data])
             self.mechanism = 'EXTERNAL'
@@ -241,8 +241,15 @@ class SASL(PlugIn):
                 self.startsasl = SASL_IN_PROCESS
                 self._owner.send(str(node))
                 raise NodeProcessed
-            except kerberos.GSSError, e:
+            except kerberos.GSSError as e:
                 log.info('GSSAPI authentication failed: %s' % str(e))
+        if 'SCRAM-SHA-1-PLUS' in self.mecs and self.channel_binding != None:
+            self.mecs.remove('SCRAM-SHA-1-PLUS')
+            self.mechanism = 'SCRAM-SHA-1-PLUS'
+            self._owner._caller.get_password(self.set_password, self.mechanism)
+            self.scram_step = 0
+            self.startsasl = SASL_IN_PROCESS
+            raise NodeProcessed
         if 'SCRAM-SHA-1' in self.mecs:
             self.mecs.remove('SCRAM-SHA-1')
             self.mechanism = 'SCRAM-SHA-1'
@@ -271,8 +278,9 @@ class SASL(PlugIn):
             self.startsasl = SASL_IN_PROCESS
             raise NodeProcessed
         self.startsasl = SASL_FAILURE
-        log.info('I can only use EXTERNAL, SCRAM-SHA-1, DIGEST-MD5, GSSAPI and '
-            'PLAIN mecanisms.')
+        log.info('I can only use ANONYMOUS, EXTERNAL, GSSAPI, SCRAM-SHA-1-PLUS,'
+                 ' SCRAM-SHA-1, DIGEST-MD5, PLAIN and X-MESSENGER-OAUTH2'
+                 ' mechanisms.')
         if self.on_sasl:
             self.on_sasl()
         return
@@ -285,10 +293,19 @@ class SASL(PlugIn):
             return
 
         def scram_base64(s):
-            return ''.join(s.encode('base64').split('\n'))
+            try:
+                s = s.encode('utf-8')
+            except:
+                pass
+            return ''.join(base64.b64encode(s).decode('utf-8').\
+                split('\n'))
 
         incoming_data = challenge.getData()
-        data=base64.decodestring(incoming_data)
+        data=base64.b64decode(incoming_data.encode('utf-8'))
+
+        if self.mechanism != 'GSSAPI':
+            data=data.decode('utf-8')
+
         ### Handle Auth result
         def on_auth_fail(reason):
             log.info('Failed SASL authentification: %s' % reason)
@@ -313,9 +330,9 @@ class SASL(PlugIn):
                 reason = challenge
             on_auth_fail(reason)
         elif challenge.getName() == 'success':
-            if self.mechanism == 'SCRAM-SHA-1':
+            if self.mechanism in ('SCRAM-SHA-1', 'SCRAM-SHA-1-PLUS'):
                 # check data-with-success
-                data = scram_parse(data)
+                data = _scram_parse(data)
                 if data['v'] != scram_base64(self.scram_ServerSignature):
                     on_auth_fail('ServerSignature is wrong')
 
@@ -339,7 +356,10 @@ class SASL(PlugIn):
             raise NodeProcessed
 
         ### Perform auth step
-        log.info('Got challenge:' + data)
+        if self.mechanism != 'GSSAPI':
+            log.info('Got challenge:' + data)
+        else:
+            log.info('Got challenge')
 
         if self.mechanism == 'GSSAPI':
             if self.gss_step == GSS_STATE_STEP:
@@ -357,15 +377,20 @@ class SASL(PlugIn):
             self._owner.send(Node('response', attrs={'xmlns': NS_SASL},
                 payload=response).__str__())
             raise NodeProcessed
-        if self.mechanism == 'SCRAM-SHA-1':
+        if self.mechanism in ('SCRAM-SHA-1', 'SCRAM-SHA-1-PLUS'):
             hashfn = hashlib.sha1
 
             def HMAC(k, s):
-                return hmac.HMAC(key=k, msg=s, digestmod=hashfn).digest()
+                return hmac.new(key=k, msg=s, digestmod=hashfn).digest()
 
-            def XOR(x, y):
-                r = (chr(ord(px) ^ ord(py)) for px, py in zip(x, y))
-                return ''.join(r)
+            if sys.version_info[0] == 2:
+                def XOR(x, y):
+                    r = (chr(ord(px) ^ ord(py)) for px, py in zip(x, y))
+                    return bytes(b''.join(r))
+            else:
+                def XOR(x, y):
+                    r = [px ^ py for px, py in zip(x, y)]
+                    return bytes(r)
 
             def Hi(s, salt, iters):
                 ii = 1
@@ -373,7 +398,7 @@ class SASL(PlugIn):
                     s = s.encode('utf-8')
                 except:
                     pass
-                ui_1 = HMAC(s, salt + '\0\0\0\01')
+                ui_1 = HMAC(s, salt + b'\0\0\0\01')
                 ui = ui_1
                 for i in range(iters - 1):
                     ii += 1
@@ -387,22 +412,31 @@ class SASL(PlugIn):
             if self.scram_step == 0:
                 self.scram_step = 1
                 self.scram_soup += ',' + data + ','
-                data = scram_parse(data)
-                # TODO: Should check cnonce here.
-                # TODO: Channel binding data goes in here too.
-                r = 'c=' + scram_base64(self.scram_gs2)
+                data = _scram_parse(data)
+                # Check server nonce here.
+                # The first part of server nonce muss be the nonce send by client.
+                if (data['r'][:len(self.client_nonce)] != self.client_nonce):
+                    on_auth_fail('Server nonce is incorrect')
+                    raise NodeProcessed
+                if self.mechanism == 'SCRAM-SHA-1':
+                    r = 'c=' + scram_base64(self.scram_gs2)
+                else:
+                    # Channel binding data goes in here too.
+                    r = 'c=' + scram_base64(self.scram_gs2
+                                             + self.channel_binding)                   
                 r += ',r=' + data['r']
                 self.scram_soup += r
-                salt = data['s'].decode('base64')
+                self.scram_soup = self.scram_soup.encode('utf-8')
+                salt = base64.b64decode(data['s'].encode('utf-8'))
                 iter = int(data['i'])
                 SaltedPassword = Hi(self.password, salt, iter)
                 # TODO: Could cache this, along with salt+iter.
-                ClientKey = HMAC(SaltedPassword, 'Client Key')
+                ClientKey = HMAC(SaltedPassword, b'Client Key')
                 StoredKey = scram_H(ClientKey)
                 ClientSignature = HMAC(StoredKey, self.scram_soup)
                 ClientProof = XOR(ClientKey, ClientSignature)
                 r += ',p=' + scram_base64(ClientProof)
-                ServerKey = HMAC(SaltedPassword, 'Server Key')
+                ServerKey = HMAC(SaltedPassword, b'Server Key')
                 self.scram_ServerSignature = HMAC(ServerKey, self.scram_soup)
                 sasl_data = scram_base64(r)
                 node = Node('response', attrs={'xmlns': NS_SASL},
@@ -411,21 +445,22 @@ class SASL(PlugIn):
                 raise NodeProcessed
 
             if self.scram_step == 1:
-                data = scram_parse(data)
-                if data['v'].decode('base64') != self.scram_ServerSignature:
+                data = _scram_parse(data)
+                if base64.b64decode(data['v'].encode('utf-8')).decode('utf-8') \
+                != self.scram_ServerSignature:
                     # TODO: Not clear what to do here - need to abort.
                     raise Exception
                 node = Node('response', attrs={'xmlns': NS_SASL});
                 self._owner.send(str(node))
                 raise NodeProcessed
 
+        # DIGEST-MD5
         # magic foo...
-        chal = challenge_splitter(data)
+        chal = _challenge_splitter(data)
         if not self.realm and 'realm' in chal:
             self.realm = chal['realm']
-        if 'qop' in chal and ((isinstance(chal['qop'], str) and \
-        chal['qop'] =='auth') or (isinstance(chal['qop'], list) and 'auth' in \
-        chal['qop'])):
+        if 'qop' in chal and ((chal['qop'] =='auth') or \
+        (isinstance(chal['qop'], list) and 'auth' in chal['qop'])):
             self.resp = {}
             self.resp['username'] = self.username
             if self.realm:
@@ -433,8 +468,7 @@ class SASL(PlugIn):
             else:
                 self.resp['realm'] = self._owner.Server
             self.resp['nonce'] = chal['nonce']
-            self.resp['cnonce'] = ''.join("%x" % randint(0, 2**28) for randint \
-                in itertools.repeat(random.randint, 7))
+            self.resp['cnonce'] = '%x' % rndg.getrandbits(196)
             self.resp['nc'] = ('00000001')
             self.resp['qop'] = 'auth'
             self.resp['digest-uri'] = 'xmpp/' + self._owner.Server
@@ -456,20 +490,27 @@ class SASL(PlugIn):
     @staticmethod
     def _convert_to_iso88591(string):
         try:
-            string = string.decode('utf-8').encode('iso-8859-1')
+            string = string.encode('iso-8859-1')
         except UnicodeEncodeError:
             pass
         return string
 
     def set_password(self, password):
         self.password = '' if password is None else password
-        if self.mechanism == 'SCRAM-SHA-1':
-            nonce = ''.join('%x' % randint(0, 2 ** 28) for randint in \
-                itertools.repeat(random.randint, 7))
-            self.scram_soup = 'n=' + self.username + ',r=' + nonce
-            self.scram_gs2 = 'n,,' # No CB yet.
-            sasl_data = (self.scram_gs2 + self.scram_soup).encode('base64').\
-                replace('\n', '')
+        if self.mechanism in ('SCRAM-SHA-1', 'SCRAM-SHA-1-PLUS'):
+            self.client_nonce = '%x' % rndg.getrandbits(196)
+            self.scram_soup = 'n=' + self.username + ',r=' + self.client_nonce
+            if self.mechanism == 'SCRAM-SHA-1':
+                if self.channel_binding == None:
+                    # Client doesn't support Channel Binding
+                    self.scram_gs2 = 'n,,' # No CB yet.
+                else:
+                    # Client supports CB, but server doesn't support CB
+                    self.scram_gs2 = 'y,,'
+            else:
+                self.scram_gs2 = 'p=tls-unique,,'
+            sasl_data = base64.b64encode((self.scram_gs2 + self.scram_soup).\
+                encode('utf-8')).decode('utf-8').replace('\n', '')
             node = Node('auth', attrs={'xmlns': NS_SASL,
                 'mechanism': self.mechanism}, payload=[sasl_data])
         elif self.mechanism == 'DIGEST-MD5':
@@ -477,30 +518,34 @@ class SASL(PlugIn):
             hash_realm = self._convert_to_iso88591(self.resp['realm'])
             hash_password = self._convert_to_iso88591(self.password)
             A1 = C([H(C([hash_username, hash_realm, hash_password])),
-                self.resp['nonce'], self.resp['cnonce']])
-            A2 = C(['AUTHENTICATE', self.resp['digest-uri']])
-            response = HH(C([HH(A1), self.resp['nonce'], self.resp['nc'],
-                self.resp['cnonce'], self.resp['qop'], HH(A2)]))
-            A2 = C(['', self.resp['digest-uri']])
-            self.digest_rspauth = HH(C([HH(A1), self.resp['nonce'],
-                self.resp['nc'], self.resp['cnonce'], self.resp['qop'],
-                HH(A2)]))
+                self.resp['nonce'].encode('utf-8'), self.resp['cnonce'].encode(
+                'utf-8')])
+            A2 = C([b'AUTHENTICATE', self.resp['digest-uri'].encode('utf-8')])
+            response = HH(C([HH(A1).encode('utf-8'), self.resp['nonce'].encode(
+                'utf-8'), self.resp['nc'].encode('utf-8'), self.resp['cnonce'].\
+                encode('utf-8'), self.resp['qop'].encode('utf-8'), HH(A2).\
+                encode('utf-8')]))
+            A2 = C([b'', self.resp['digest-uri'].encode('utf-8')])
+            self.digest_rspauth = HH(C([HH(A1).encode('utf-8'), self.resp[
+                'nonce'].encode('utf-8'), self.resp['nc'].encode('utf-8'),
+                self.resp['cnonce'].encode('utf-8'), self.resp['qop'].encode(
+                'utf-8'), HH(A2).encode('utf-8')]))
             self.resp['response'] = response
-            sasl_data = u''
+            sasl_data = ''
             for key in ('charset', 'username', 'realm', 'nonce', 'nc', 'cnonce',
             'digest-uri', 'response', 'qop'):
                 if key in ('nc', 'qop', 'response', 'charset'):
-                    sasl_data += u"%s=%s," % (key, self.resp[key])
+                    sasl_data += "%s=%s," % (key, self.resp[key])
                 else:
-                    sasl_data += u'%s="%s",' % (key, self.resp[key])
-            sasl_data = sasl_data[:-1].encode('utf-8').encode('base64').replace(
-                '\r', '').replace('\n', '')
+                    sasl_data += '%s="%s",' % (key, self.resp[key])
+            sasl_data = base64.b64encode(sasl_data[:-1].encode('utf-8')).\
+                decode('utf-8').replace('\r', '').replace('\n', '')
             node = Node('response', attrs={'xmlns': NS_SASL},
                 payload=[sasl_data])
         elif self.mechanism == 'PLAIN':
-            sasl_data = u'\x00%s\x00%s' % (self.username, self.password)
-            sasl_data = sasl_data.encode('utf-8').encode('base64').replace(
-                '\n', '')
+            sasl_data = '\x00%s\x00%s' % (self.username, self.password)
+            sasl_data = base64.b64encode(sasl_data.encode('utf-8')).decode(
+                'utf-8').replace('\n', '')
             node = Node('auth', attrs={'xmlns': NS_SASL, 'mechanism': 'PLAIN'},
                 payload=[sasl_data])
         elif self.mechanism == 'X-MESSENGER-OAUTH2':
@@ -512,7 +557,7 @@ class SASL(PlugIn):
 
 class NonBlockingNonSASL(PlugIn):
     """
-    Implements old Non-SASL (JEP-0078) authentication used in jabberd1.4 and
+    Implements old Non-SASL (XEP-0078) authentication used in jabberd1.4 and
     transport authentication
     """
 
@@ -575,7 +620,7 @@ class NonBlockingNonSASL(PlugIn):
             query.setTagData('hash', hash_)
             self._method='0k'
         else:
-            log.warn("Secure methods unsupported, performing plain text \
+            log.warning("Secure methods unsupported, performing plain text \
                 authentication")
             self._method = 'plain'
             self._owner._caller.get_password(self._on_password, self._method)
