@@ -472,7 +472,8 @@ class NonBlockingTCP(NonBlockingTransport, IdleObject):
             self._sock.shutdown(socket.SHUT_RDWR)
             self._sock.close()
         except socket.error as e:
-            log.info('Error while disconnecting socket: %s' % str(e))
+            errstr = e.strerror
+            log.info('Error while disconnecting socket: %s' % errstr)
         self.fd = -1
         NonBlockingTransport.disconnect(self, do_callback)
 
@@ -555,7 +556,7 @@ class NonBlockingTCP(NonBlockingTransport, IdleObject):
                 return None
             self.sendbuff = self.sendqueue.pop(0)
         try:
-            send_count = self._send(self.sendbuff.encode("utf-8"))
+            send_count = self._send(self.sendbuff)
             if send_count:
                 sent_data = self.sendbuff[:send_count]
                 self.sendbuff = self.sendbuff[send_count:]
@@ -762,9 +763,7 @@ class NonBlockingHTTP(NonBlockingTCP):
         Builds http message with given body. Values for headers and status line
         fields are taken from class variables
         """
-        absolute_uri = '%s://%s:%s%s' % (self.http_protocol, self.http_host,
-                self.http_port, self.http_path)
-        headers = ['%s %s %s' % (method, absolute_uri, self.http_version),
+        headers = ['%s %s %s' % (method, self.http_path, self.http_version),
                 'Host: %s:%s' % (self.http_host, self.http_port),
                 'User-Agent: Gajim',
                 'Content-Type: text/xml; charset=utf-8',
